#!/usr/bin/python3

MIN_I16 = -0x8000
MAX_I16 = 0x7fff
MAX_UI16 = 0xffff

MIN_I32 = -0x80000000
MAX_I32 = 0x7fffffff
MAX_UI32 = 0xffffffff

MIN_I64 = -0x8000000000000000
MAX_I64 = 0x7fffffffffffffff
MAX_UI64 = 0xffffffffffffffff
