#!/usr/bin/python3
# encoding=utf-8


g_merk_nack_codes = {
    1: 'Недопустимая команда или параметр',
    2: 'Внутренняя ошибка счетчика',
    3: 'Не достаточен уровень доступа для удовлетворения запроса',
    4: 'Внутренние часы счетчика уже корректировались в течение текущих суток',
    5: 'Не открыт канал связи',
}

g_merk_type_defs = {
    'merk230': {},
}
